# -*- coding: utf-8 -*-
"""Parser for Trend Micro Antivirus logs.

Trend Micro uses two log files to track the scans (both manual/scheduled and
real-time) and the web reputation (network scan/filtering).

Currently only the first log is supported.
"""

from __future__ import unicode_literals

from plaso.containers import events
from plaso.containers import time_events
from plaso.lib import errors
from plaso.lib import definitions
from plaso.lib import timelib
from plaso.parsers import dsv_parser
from plaso.parsers import manager

_SCAN_RESULTS = {
    0: "Success (clean)",
    1: "Success (move)",
    2: "Success (delete)",
    3: "Success (rename)",
    4: "Pass > Deny access",
    5: "Failure (clean)",
    6: "Falure (move)",
    7: "Failure (delete)",
    8: "Failure (rename)",
    10: "Failure (clean), moved",
    11: "Failure (clean), deleted",
    12: "Failure (clean), renamed",
    13: "Pass > Deny access",
    14: "Failure (clean), move also failed",
    15: "Failure (clean), delete also failed",
    16: "Failure (clean), rename also failed",
    25: "Passed a potential security risk"
}

_SCAN_TYPES = {
    0: "Manual scan",
    1: "Real-time scan",
    2: "Scheduled scan",
    3: "Scan Now scan",
    4: "DCS scan"
}

_BLOCK_MODES = {
    0: "Internal filter",
    1: "Whitelist only"
}


class TrendMicroAVEventData(events.EventData):
  """Trend Micro AV Log event data.

  Attributes:
    action (str): action.
    threat (str): threat.
    filename (str): filename.
    scan_type (str): scan_type.
    trigger_location (str): trigger location.
    username (str): username.
  """

  DATA_TYPE = 'av:trendmicro:scan'

  def __init__(self):
    """Initializes event data."""
    super(TrendMicroAVEventData, self).__init__(data_type=self.DATA_TYPE)
    self.threat = None
    self.action = None
    self.path = None
    self.filename = None
    self.scan_type = None


# pylint: disable=abstract-method
class TrendMicroBaseParser(dsv_parser.DSVParser):
  """Common code for parsing Trend Micro log files."""
  DELIMITER = '<;>'

  # Subclasses must define an integer MIN_COLUMNS value.
  MIN_COLUMNS = None

  # Subclasses must define a list of field names.
  COLUMNS = ()

  def __init__(self, *args, **kwargs):
    kwargs['encoding'] = 'cp1252'
    super(TrendMicroBaseParser, self).__init__(*args, **kwargs)

  def _CreateDictReader(self, parser_mediator, line_reader):
    """Iterate over the log lines and provide a reader for the values.

    The file format is reminiscent of CSV, but is not quite the same; the
    delimiter is a three-character sequence and there is no provision for
    quoting or escaping.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      line_reader (iter): yields each line in the log file.
    """
    for line in line_reader:
      try:
        values = line.decode(self._encoding).strip().split(self.DELIMITER)
      except UnicodeDecodeError, exception:
        raise errors.UnableToParseFile(
            "Unexpected binary content in file: {0:s}".format(exception))
      if len(values) < self.MIN_COLUMNS:
        raise errors.UnableToParseFile(
            "Expected at least {0:d} values, found {1:d}".format(
                self.MIN_COLUMNS, len(values)))
      if len(values) > len(self.COLUMNS):
        raise errors.UnableToParseFile(
            "Expected at most {0:d} values, found {1:d}".format(
                len(self.COLUMNS), len(values)))
      yield dict(zip(self.COLUMNS, values))

  def _ParseTimestamp(self, parser_mediator, row):
    """Provides a timestamp for the given row.

    If the Trend Micro log comes from a version that provides a Unix timestamp,
    use that directly; it provides the advantages of UTC and of second
    precision.  Otherwise fall back onto the local-timezone date and time.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row (dict[str, str]): fields of a single row, as specified in COLUMNS.

    Returns:
      int: a timestamp integer containing the number of micro seconds since
          January 1, 1970, 00:00:00 UTC.
    """
    if 'timestamp' in row:
      try:
        return timelib.Timestamp.FromPosixTime(int(row['timestamp']))
      except errors.TimestampError as exception:
        parser_mediator.ProduceExtractionError(
            'Log line has a timestamp field: [{0:s}], but it is invalid: {1:s}'
            .format(repr(row['timestamp']), exception))
    else:
      # The Unix timestamp is not available; parse the local date and time.
      try:
        return self._ConvertToTimestamp(
            row['date'], row['time'], parser_mediator.timezone)
      except errors.TimestampError as exception:
        parser_mediator.ProduceExtractionError(
            'Unable to parse time string: [{0:s} {1:s}] with error {2:s}'
            .format(repr(row['date']), repr(row['time']), exception))


  def _ConvertToTimestamp(self, date, time, timezone):
    """Converts date and time strings into a timestamp.

    Recent versions of Office Scan write a log field with a Unix timestamp.
    Older versions may not write this field; their logs only provide a date and
    a time expressed in the local time zone. This functions handles the latter
    case.

    The date value is an 8-character string in the YYYYMMDD format.

    The time value may be a 4-character string in the HHMM format or a
    6-character string in the HHMMSS format.

    Args:
      timestamp (str): Unix time in seconds since the epoch.
      date (str): date.
      time (str): time.
      timezone (pytz.timezone): timezone of the date and time.

    Returns:
      int: a timestamp integer containing the number of micro seconds since
          January 1, 1970, 00:00:00 UTC.

    Raises:
      TimestampError: if the timestamp is badly formed or unable to transfer
          the supplied date and time into a timestamp.
    """
    year = int(date[:4])
    month = int(date[4:6])
    day = int(date[6:8])
    hour = int(time[:2])
    minutes = int(time[2:4])
    seconds = 0
    if len(time) > 4:
      seconds = time[4:6]
    return timelib.Timestamp.FromTimeParts(
        year, month, day, hour, minutes, seconds, timezone=timezone)

  def _VerifyLengthAndTimestamp(self, parser_mediator, row):
    """Verifies the number of elements and the timestamp in the row.

    Checks that the number of elements matches the format described by
    COLUMNS and MIN_COLUMNS, and that the 'date' and 'time' fields
    express a plausible timestamp.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row (dict[str, str]): fields of a single row, as specified in COLUMNS.

    Returns:
      bool: True if this may be the correct parser, False otherwise.
    """
    field_count = len(row)
    if field_count < self.MIN_COLUMNS or field_count > len(self.COLUMNS):
      return False

    # Check the date format!
    # If it doesn't parse, then this isn't a Trend Micro AV log.
    timestamp = self._ConvertToTimestamp(
        row['date'], row['time'], parser_mediator.timezone)

    return timestamp is not None


class OfficeScanVirusDetectionParser(TrendMicroBaseParser):
  """Parses the Trend Micro Office Scan Virus Detection Log."""

  NAME = 'trendmicro_vd'
  DESCRIPTION = 'Parser for Trend Micro Office Scan Virus Detection log files.'

  COLUMNS = (
      'date', 'time', 'threat', 'action', 'scan_type', 'unused1',
      'path', 'filename', 'unused2', 'timestamp', 'unused3', 'unused4')
  MIN_COLUMNS = 8

  def ParseRow(self, parser_mediator, row_offset, row):
    """Parses a line of the log file and produces events.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row_offset (int): line number of the row.
      row (dict[str, str]): fields of a single row, as specified in COLUMNS.
    """

    timestamp = self._ParseTimestamp(parser_mediator, row)

    if timestamp is None:
      return

    event_data = TrendMicroAVEventData()
    event_data.offset = row_offset
    event_data.threat = row['threat']
    event_data.action = _SCAN_RESULTS[int(row['action'])]
    event_data.path = row['path']
    event_data.filename = row['filename']
    event_data.scan_type = _SCAN_TYPES[int(row['scan_type'])]

    event = time_events.TimestampEvent(
        timestamp, definitions.TIME_DESCRIPTION_WRITTEN)
    parser_mediator.ProduceEventWithEventData(event, event_data)

  def VerifyRow(self, parser_mediator, row):
    """Verifies if a line of the file is in the expected format.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row (dict[str, str]): fields of a single row, as specified in COLUMNS.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    if not self._VerifyLengthAndTimestamp(parser_mediator, row):
      return False

    # Check that the action value is plausible
    try:
      action = int(row['action'])
    except ValueError:
      return False
    if action not in _SCAN_RESULTS:
      return False

    # All checks passed.
    return True


class TrendMicroUrlEventData(events.EventData):
  """Trend Micro Web Reputation Log event data.

  Attributes:
    block_mode (str): operation mode.
    url (str): accessed URL.
    group_code (str): group code.
    group_name (str): group name.
    cred_rating (int): credibility rating.
    cred_score (int): credibility score.
    policy_id (int): policy ID.
    appname (str): application name.
    ip (str): IP address.
    threshold (int): threshold value.
  """
  DATA_TYPE = 'av:trendmicro:webrep'

  def __init__(self):
    """Initializes event data."""
    super(TrendMicroUrlEventData, self).__init__(data_type=self.DATA_TYPE)
    self.block_mode = None
    self.url = None
    self.group_code = None
    self.group_name = None
    self.cred_rating = None
    self.cred_score = None
    self.policy_id = None
    self.appname = None
    self.ip = None
    self.threshold = None


class OfficeScanWebReputationParser(TrendMicroBaseParser):
  """Parses the Trend Micro Office Scan Web Reputation detection log."""
  NAME = 'trendmicro_url'
  DESCRIPTION = 'Parser for Trend Micro Office Web Reputation log files.'

  COLUMNS = (
      'date', 'time', 'block_mode', 'url', 'group_code', 'group_name',
      'cred_rating', 'policy_id', 'appname', 'cred_score', 'ip', 'threshold',
      'timestamp', 'unused')

  MIN_COLUMNS = 12

  def ParseRow(self, parser_mediator, row_offset, row):
    """Parses a line of the log file and produces events.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row_offset (int): line number of the row.
      row (dict[str, str]): fields of a single row, as specified in COLUMNS.
    """

    timestamp = self._ParseTimestamp(parser_mediator, row)

    if timestamp is None:
      return

    event_data = TrendMicroUrlEventData()
    event_data.offset = row_offset

    # Parse and store block mode.
    event_data.block_mode = _BLOCK_MODES[int(row['block_mode'])]

    # Convert and store integer values.
    for field in ('cred_rating', 'cred_score', 'policy_id', 'threshold'):
      setattr(event_data, field, int(row[field]))

    # Store string values.
    for field in ('url', 'group_name', 'group_code', 'appname', 'ip'):
      setattr(event_data, field, row[field])

    event = time_events.TimestampEvent(
        timestamp, definitions.TIME_DESCRIPTION_WRITTEN)
    parser_mediator.ProduceEventWithEventData(event, event_data)

  def VerifyRow(self, parser_mediator, row):
    """Verifies if a line of the file is in the expected format.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row (dict[str, str]): fields of a single row, as specified in COLUMNS.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    if not self._VerifyLengthAndTimestamp(parser_mediator, row):
      return False

    # Check the block mode.
    try:
      if int(row['block_mode']) not in _BLOCK_MODES:
        return False
    except ValueError:
      return False

    return True


manager.ParsersManager.RegisterParser(OfficeScanVirusDetectionParser)
manager.ParsersManager.RegisterParser(OfficeScanWebReputationParser)
